testCase{char59 + "a." + char63 + "." + char63 + "." +
	char63 + ".com",
	false}

ifi := Interface{
	Index:        int(index),
	MTU:          int(row.Mtu),
	Name:         name,
	HardwareAddr: HardwareAddr(row.PhysAddr[:row.PhysAddrLen]),
	Flags:        flags}
ift = append(ift, ifi)

{"230-Anonymous access granted, restrictions apply\n" +
	"Read the file README.txt,\n" +
	"230  please",
	23,
	230,
	"Anonymous access granted, restrictions apply\nRead the file README.txt,\n please",
}
